# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from django.contrib import auth
from django.contrib.auth import get_user_model

from m3_auth.utils import import_object
from m3_auth.utils import get_request_params
from m3_auth.exceptions import AuthError
from m3_auth.backends.base import BaseBackend
from m3_auth.backends.django.checkers import DefaultChecker


class DjangoBackend(BaseBackend):

    """
    Бэкенд, реализующий аутентификацию по логину и паролю средствами django.
    
    :param name: Наименование бэкенда.
    :param config: Настройки бэкенда, ожидается объект класса Config с 
        параметрами login_url и logout_url.
    :param checker: Класс для проверки данных пользователя, по умолчанию 
        используется класс DefaultChecker.    
    """

    def __init__(self, name, config, checker=DefaultChecker):
        super(DjangoBackend, self).__init__(name)
        self.config = config
        self.checker = checker

    def get_login_url(self, request):
        """Возвращает URL адрес на который необходимо отправить запрос с 
        данными для аутентификации. URL адрес берется из настроек.
        
        :param request: Запрос.
        :return: URL адрес.
        :rtype: str
        """
        return self.config.login_url

    def process_login(self, request):
        """Метод обрабатывает запрос на аутентификацию. 
        
        В параметрах запроса ожидается:
        1. username с логином пользователя, либо USERNAME_FIELD из настроек 
        модели пользователя, например, может быть email.
        2. password - пароль пользователя.
        
        В случае возникновения ошибок, возникает исключение AuthError с 
        сообщением об ошибке.
        
        :param request: Запрос.
        :type request: django.http.HttpRequest
        :raises: auth.exceptions.AuthError
        :return: Объект класса пользователя User.
        :rtype: django.contrib.auth.models.AbstractUser
        """
        user_model = get_user_model()
        model_meta = user_model._meta
        # Аутентификация может быть и по email'у и паролю.
        username_field = getattr(user_model, 'USERNAME_FIELD', 'username')

        params = get_request_params(request)
        username = params.get(username_field)
        password = params.get('password')

        if not all((username, password)):
            raise AuthError(
                'Отсутствуют обязательные параметры: "{}" и/или "{}".'.format(
                    model_meta.get_field(username_field).verbose_name,
                    model_meta.get_field('password').verbose_name
                ))

        error = self.checker(request, username, password).check()
        if error:
            raise AuthError(error)

        user = auth.authenticate(username=username, password=password)
        auth.login(request, user)
        return user

    def get_logout_url(self, request):
        """"Возвращает URL адрес на который необходимо отправить запрос для 
        выхода из системы. URL адрес берется из настроек.
        
        :param request: Запрос.
        :type request: django.http.HttpRequest
        :return: URL адрес.
        :rtype: str
        """
        return self.config.logout_url

    def process_logout(self, request):
        """Метод осуществляет выход из системы. 
        
        :param request: Запрос.
        :type request: django.http.HttpRequest
        """
        auth.logout(request)

    @classmethod
    def from_config(cls, config):
        """Метод создает и возвращает экземпляр класса бэкенда на основе 
        параметров config.
        
        В конфиге должны быть два обязательных параметра:
        1. name - наименование бэкенда.
        2. config - Конфиги используемые самим бэкендом.
        
        и третий не обязательный параметр checker_cls - строка с ссылкой на 
        класс (н-р. auth.backends.django.checker.DefaultChecker)
        для проверки данных пользователя в момент обработки запроса на 
        аутентификацию. По умолчанию используется DefaultChecker.
        
        :param config: Конфигурация бэкенда.
        :type config: auth.config.Config
        :return: Экземпляр класса cls.
        :rtype: auth.backends.django.DjangoBackend
        """
        checker_cls = DefaultChecker
        if config.checker_cls:
            checker_cls = import_object(config.checker_cls)

        return cls(config.name, config.config, checker_cls)
