# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from uuid import uuid4
from datetime import datetime

import pytz
import requests
from oauthlib.common import urldecode
from requests.auth import HTTPBasicAuth
from oauthlib.oauth2 import WebApplicationClient

from m3_auth.config import Config
from m3_auth.utils import filter_params
from m3_auth.utils import import_object
from m3_auth.utils import get_request_params
from m3_auth.exceptions import AuthError
from m3_auth.utils import add_query_params
from m3_auth.backends.base.remote.protocol import BaseProtocol
from m3_auth.backends.oauth2.exceptions import RefreshTokenError


class OAuthProtocol(BaseProtocol):

    """Класс реализует формирование URL адресов, обработку запросов на 
    аутентификацию, получение и обновление токена по протоколу OAuth 2.0.
    
    :param user_cls: Класс который будет представлять информацию о пользователе 
        с сервера авторизации.
    :type user_cls: type

    :param api: Объект класса реализующий получения данных о пользователе с
        сервера авторизации.
    :type api: auth.backends.oauth2.api.BaseOAuthServerApi

    :param config: Настройки протокола.
    :type config: auth.config.Config
    """

    def __init__(self, user_cls, api, config):
        super(OAuthProtocol, self).__init__(user_cls, config)
        self.signer = None
        self.api = api

        # Если передан объект а не строка, значит будет использоваться подпись
        if isinstance(config.client_secret, Config):
            self.signer = self._load_signer(config.client_secret)

    def get_login_url(self, request):
        """Метод формирует и возвращает URL адрес для переадресации на сервер 
        авторизации.
        
        Список передаваемых параметров можно настраивать настройкой
        config.login.request_params. Есть 3 доступных параметра: client_secret,
        timestamp, state.
        
        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        
        :return: URL адрес.
        :rtype: str
        """
        state, timestamp, secret = self._get_request_params(request)
        login_params = filter_params({
            'client_secret': secret,
            'timestamp': timestamp,
            'state': state
        }, self.config.login.request_params)

        return self._get_client().prepare_request_uri(
            self.config.login.url,
            redirect_uri=self.config.login.redirect_url,
            scope=self.config.scope,
            **login_params)

    def process_login(self, request):
        """Метод обрабатывает запрос на вход, запрашивает токен с сервера 
        авторизации, запрашивает информацию о пользователе и возвращает объект 
        класса user_cls с информацией о пользователе.

        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        :raises: auth.exceptions.AuthError

        :return: Объект пользователя.
        :rtype: auth.backends.base.remote.user.BaseRemoteUser
        """
        params = get_request_params(request)
        if 'error' in params and 'error_description' in params:
            raise AuthError(params['error_description'])

        token = self.get_token(request)
        client = self._get_client(token=token)
        person_info = self.api.get_person_info(request, client)
        return self.user_cls(person_info)

    def get_logout_url(self, request):
        """Возвращает URL адрес для переадресации пользователя на выход.
        
        Список передаваемых параметров можно настраивать настройкой
        config.logout.request_params. Есть 2 доступных параметра: redirect_url
        и client_id.
        
        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        :return: URL адрес.
        :rtype: str
        """
        logout_params = filter_params({
            'redirect_url': self.config.logout.redirect_url,
            'client_id': self.config.client_id
        }, self.config.logout.request_params or [])

        return add_query_params(self.config.logout.url, logout_params)

    def process_logout(self, request):
        """Метод просто возвращает True, т.к. обрабатывать в принципе нечего.
        Обычно это редирект с сервера авторизации на URL реализующий выход.

        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        :return: True
        :rtype: bool
        """
        return True

    def refresh_token(self, request):
        """Метод отправляет запрос на сервер авторизации для обновления токена.
        Запрос отправится только в том случае, если сервер авторизации вернул 
        токен с параметром refresh_token. 
        
        Список передаваемых параметров можно настраивать настройкой
        config.token.request_params. Есть 3 доступных параметра: client_secret,
        timestamp и state.
        
        Метод выкидывает исключение RefreshTokenError если токен был отозван 
        или устарел.
        
        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        :raises: auth.protocols.oauth.exceptions.RefreshTokenError
        
        :return: True
        :rtype: bool
        """
        token = self._get_oauth_token(request)
        if 'refresh_token' not in token:
            return None

        state, timestamp, secret = self._get_request_params(request)
        token_params = filter_params({
            'client_secret': secret,
            'timestamp': timestamp,
            'state': state
        }, self.config.token.request_params)

        body = self._get_client(token).prepare_request_body(
            redirect_uri=self.config.token.redirect_url,
            refresh_token=token['refresh_token'],
            scope=self.config.scope,
            **token_params)

        data = dict(urldecode(body))
        data['grant_type'] = 'refresh_token'

        token_request = requests.post(
            self.config.token.url,
            data=data,
            verify=self.config.ssl_verification,
            headers=self._get_headers())

        if not token_request.ok:
            raise RefreshTokenError('Ошибка обновления токена.')

        self._save_oauth_token(request, token_request.json())

    def get_token(self, request):
        """Метод формирует от отправляет запрос на получения токена с сервера 
        авторизации.
        
        Список передаваемых параметров можно настраивать настройкой
        config.token.request_params. Есть 3 доступных параметра: client_secret,
        timestamp и state.
        
        В случае ответа с ошибкой, будет выброшено исключение AuthError.
        
        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        :raises: auth.exceptions.AuthError
        
        :return: OAuth токен.
        :rtype: dict
        """

        params = get_request_params(request)
        state, timestamp, secret = self._get_request_params(request)
        auth = HTTPBasicAuth(self.config.client_id, secret)

        token_params = filter_params({
            'client_secret': secret,
            'timestamp': timestamp,
            'state': state
        }, self.config.token.request_params)

        # Формируем данные для запроса.
        body = self._get_client().prepare_request_body(
            redirect_uri=self.config.token.redirect_url,
            scope=self.config.scope,
            code=params['code'],
            **token_params)

        token_request = requests.post(
            self.config.token.url,
            data=dict(urldecode(body)),
            auth=auth,
            verify=self.config.ssl_verification,
            headers=self._get_headers())

        if not token_request.ok:
            raise AuthError('Ошибка получения токена.')

        token = token_request.json()
        self._save_oauth_token(request, token)
        return token

    def _get_client(self, token=None):
        """Метод подготавливает клиент и возвращает его.

        :param token: OAuth токен.
        :type token: dict
        
        :return: Клиент.
        :rtype: WebApplicationClient
        """
        return WebApplicationClient(
            self.config.client_id,
            token=token or {})

    def _get_client_secret(self, params):
        """Метод возвращает пароль, необходимый для сервера авторизации. 
        Это может быть обычная строка в конфиге, либо результат работы вызова
        объекта signer.

        :param params: Параметры запроса, нужны для подписывания.
        :type params: dict
        :return: Пароль.
        :rtype: str
        """
        secret = self.config.client_secret
        return self.signer(params) if self.signer else secret

    def _get_request_params(self, request):
        """Метод возвращает необходимые параметры для всех видов запросов.

        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        :return: state (для защиты от CSRF), время, пароль
        :rtype: tuple
        """
        state = self._get_state(request) or self._set_state(request)
        timestamp = self._get_timestamp()
        secret = self._get_client_secret({
            'client_id': self.config.client_id,
            'scope': ' '.join(self.config.scope),
            'timestamp': timestamp,
            'state': state
        })
        return state, timestamp, secret

    @staticmethod
    def _set_state(request, state=None):
        """Метод устанавливает и сохраняет значение OAuth state, для защиты от
        CSRF.

        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        
        :param state: Рандомное значение, для защиты от CSRF.
        :type state: str or unicode.
        
        :return: state
        :rtype: str
        """
        state = state or str(uuid4())
        request.session['oauth_state'] = state
        return state

    @staticmethod
    def _get_state(request):
        """Возвращает сохраненное значение state.

        :param request: Запрос.
        :type request: django.http.request.HttpRequest
        
        :return: OAuth state
        :rtype: str or None
        """
        return request.session.get('oauth_state')

    @staticmethod
    def _get_timestamp():
        """
        :return: Возвращает дату и время.
        :rtype: str
        """
        return datetime.now(pytz.utc).strftime(
            '%Y.%m.%d %H:%M:%S %z').strip()

    @staticmethod
    def _get_headers():
        """
        :return: Необходимые заголовки для запросов.
        :rtype: dict
        """
        return {
            'Accept': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8'
        }

    @staticmethod
    def _load_signer(config):
        """Метод импортирует класс подписчика и возвращает его объект.
        
        :param config: Конфиг для подписчика.
        :type config: auth.config.Config
        
        :return: Объект класса.
        """
        signer = import_object(config.signer_cls)
        return signer(**config.params)

    @staticmethod
    def _save_oauth_token(request, token):
        """Метод сохраняет токен в сессии.
        
        :param request: Запрос.
        :param token: Токен.
        :return: None
        """
        request.session['oauth_token'] = token

    @staticmethod
    def _get_oauth_token(request):
        """Метод возвращает сохраненный токен из сессии.
        
        :param request: Запрос.
        :return: Токен.
        """
        return request.session.get('oauth_token')

    @classmethod
    def from_config(cls, config):
        """Создает объект протокола на основе config.
        Информацию о ожидаемых параметров конфигурации можно найти в 
        документации.
        
        :param config: Настройки протокола.
        :return: Объект протокола.
        :rtype: cls
        """
        user_class = import_object(config.user_cls)
        api_class = import_object(config.api_cls)
        api_object = api_class(config.api_config)

        return cls(user_class, api_object, config.config)
