# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from importlib import import_module

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured

from m3_auth.backends.base.remote import RemoteBackend
from m3_auth.utils import check_dependency


if not check_dependency('python3-saml', '1.10.1.1'):
    raise ImproperlyConfigured(
        'SAMLBackend depends on python3-saml==1.10.1.1.'
        'You need to check your environment.'
    )


class SAMLBackend(RemoteBackend):

    """
    Класс бэкенда, для аутентификации по протоколу SAML2. 
    Дополнительно реализует Singe Logout.    
    """

    def process_single_logout(self, request):
        """Метод передает запрос протоколу на обработку, и получает 
        идентификатор пользователя, далее для этого пользователя удаляется 
        сессияя, тем самым осуществляется выход.
        
        :param request: Запрос.
        :return: 
        """
        user_id, response_url = self.protocol.process_single_logout(request)
        self._force_logout(user_id)
        return response_url

    def _login(self, request, user, remote_user_id):
        """Аутентификация пользователя в системе без пароля.
        Метод симулирует работу django.contrib.auth.authenticate().
        При этом, в кэше сохраняет идентификатор пользователя сервера 
        авторизации для возможности реализации Single Logout.

        :param request: Запрос
        :type request: django.http.request.HttpRequest 

        :param user: Пользователь системы.
        :type user: django.contrib.auth.models.User

        :param remote_user_id: Идентификатор пользователя с сервера авторизации        
        :return: None
        """
        cache_key = self._get_cache_key(remote_user_id)
        cache.delete(cache_key)
        super(SAMLBackend, self)._login(request, user, remote_user_id)
        cache.set(cache_key, request.session.session_key)

    def _force_logout(self, remote_user_id):
        """Метод удаляет сессию пользователя, тем самым делая его не 
        аутентифицированным.

        :param remote_user_id: Идентификатор пользователя сервера авторизации.
        :type remote_user_id: int
        :return: None
        """
        cache_key = self._get_cache_key(remote_user_id)
        session_key = cache.get(cache_key)
        cache.delete(cache_key)
        session_store = import_module(settings.SESSION_ENGINE).SessionStore
        session_store(session_key).delete()

    def _get_cache_key(self, user_id):
        """
        :param user_id: Идентификатор пользователя
        :type user_id: int
        :return: Ключ кэша, в которм будет хранится информация о пользователе.
        :rtype: str
        """
        return ':'.join((self.name, 'user_id', str(user_id)))
