# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from collections import OrderedDict
from importlib import import_module

from six.moves.urllib.parse import urlencode
from six.moves.urllib.parse import urlsplit
from six.moves.urllib.parse import parse_qs
from six.moves.urllib.parse import urlunsplit

from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution
from pkg_resources.extern.packaging.version import LegacyVersion
from pkg_resources.extern.packaging.version import Version
from pkg_resources.extern.packaging.version import Infinity


class SingletonMeta(type):

    """Метакласс для классов-одиночек."""

    def __init__(cls, *args, **kwargs):
        super(SingletonMeta, cls).__init__(*args, **kwargs)
        cls._instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instance


def import_object(reference):
    """Функция импортирует объект из модуля и возвращает его.
    
    :param reference: Путь до объекта. (н-р, auth.utils.import_object)
    :type reference: str or None.
    :return: Импортируемый объект.
    """
    module_name, class_name = reference.rsplit('.', 1)
    module_object = import_module(module_name)
    return getattr(module_object, class_name)


def add_query_params(url, params):
    """Добавляет параметры к URL адресу.
    
    :param url: URL адрес.
    :type url: str or unicode
    
    :param params: Словарь параметров для добавления.
    :type params: dict
    
    :return: Новая строка URL с параметрами params.
    :rtype: str or unicode
    
    :Examples:
    
    >>> url = 'http://localhost:8000/'
    >>> add_query_params(url, {'backend': 'django', 'json': 'true'})
    http://localhost:8000/?backend=django&json=true
    """

    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = OrderedDict(parse_qs(query_string))

    for param_name, param_value in sorted(params.items()):
        query_params[param_name] = [param_value]

    new_query_string = urlencode(query_params, doseq=True)
    return urlunsplit((scheme, netloc, path, new_query_string, fragment))


def filter_params(params, keys):
    """Фильтрует словарь params оставляя только те элементы, ключи которых
    находятся в списке keys.

    :param params: Словарь, который необходимо отфильтровать.
    :type params: dict
    
    :param keys: Список ключей, которые должны быть на выходе.
    :type keys: Iterable
    
    :return: Новый словарь.
    :rtype: dict
    """
    return {k: v for k, v in params.items() if k in keys}


def get_request_params(request):
    """Возвращает параметры HTTP-запроса вне зависимости от его типа."""
    result = {}
    if request.method == 'GET':
        result = request.GET
    elif request.method == 'POST':
        result = request.POST

    return result


def is_authenticated(request):
    """
    Метод проверяет, аутентифицирован пользователь или нет.
    Используется для совместимости с django 2.0.
    """
    result = request.user.is_authenticated
    return result() if callable(result) else result


def check_dependency(name, min_version, max_version=None):
    """Проверяет зависимости.

    :param name: Наименование дистрибутива.
    :type name: str

    :param min_version: Минимальная версия дистрибутива.
    :type min_version: str

    :param max_version: Максимальная версия дистрибутива.
    :type max_version: str or None
    """
    result = False

    try:
        dist = get_distribution(name)
    except DistributionNotFound:
        pass
    else:
        result = (
            not isinstance(dist.parsed_version, LegacyVersion) and
            Version(min_version) <= dist.parsed_version <
            (Version(max_version) if max_version else Infinity)
        )

    return result
