# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.contrib.auth.middleware import AuthenticationMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.test.client import RequestFactory
from django.test.testcases import TestCase

from m3_auth.backends.django import DjangoBackend
from m3_auth.backends.django.checkers import DefaultChecker
from m3_auth.config import Config
from m3_auth.exceptions import AuthError
from m3_auth.utils import is_authenticated


class CustomChecker(DefaultChecker):
    """Для проверки установки нового чекера."""


class DjangoBackendTestCase(TestCase):

    """Проверка бэкенда django."""

    def setUp(self):
        super(DjangoBackendTestCase, self).setUp()
        self.factory = RequestFactory()
        self.login_url = 'http://localhost:8000/process_login'
        self.logout_url = 'http://localhost:8000/process_logout'
        self.config = Config({
            'login_url': self.login_url,
            'logout_url': self.logout_url
        })
        self.backend = DjangoBackend('django', self.config)

        self.admin_password = 'admin'
        self.user_admin = User.objects.get(username='admin')

        # Не активный пользователь
        self.inactive_password = 'inactive'
        self.inactive_user = User.objects.get(username='inactive')

    def test_get_login_url(self):
        """Проверка метода get_login_url."""
        request = self.factory.get('/')
        login_url = self.backend.get_login_url(request)
        self.assertEqual(login_url, self.login_url)

    def test_process_login(self):
        """Проверка аутентификации."""

        request = self.factory.post('/')
        # Если не переданы обязательные параметры, должно быть исключение.
        self.assertRaises(AuthError, self.backend.process_login, request)

        # Проверка входа с некорректными данными
        request = self.factory.post('/', {
            'username': 'test',
            'password': 'test'
        })
        self.assertRaises(AuthError, self.backend.process_login, request)

        # Проверка входа с некорректным паролем
        request = self.factory.post('/', {
            'username': self.user_admin.username,
            'password': 'password'
        })
        self.assertRaises(AuthError, self.backend.process_login, request)

        # Проверка входа под не активным пользователем
        request = self.factory.post('/', {
            'username': self.inactive_user.username,
            'password': self.inactive_password
        })
        self.assertRaises(AuthError, self.backend.process_login, request)

        # Проверка аутентификации под корректными данными
        request = self.factory.post('/', {
            'username': self.user_admin.username,
            'password': self.admin_password
        })
        self.apply_middleware(request)
        user = self.backend.process_login(request)

        self.assertEqual(user.username, 'admin')
        self.assertIsInstance(user, User)
        self.assertTrue(is_authenticated(request))

    def test_process_login_with_custom_username_field(self):
        """Проверка входа, если USERNAME_FIELD в модели User не равна 
        username."""

        default_username_field = User.USERNAME_FIELD
        User.USERNAME_FIELD = 'email'

        request = self.factory.post('/', {
            'email': self.user_admin.email,
            'password': self.admin_password
        })
        self.apply_middleware(request)

        user = self.backend.process_login(request)
        User.USERNAME_FIELD = default_username_field

        self.assertTrue(is_authenticated(request))
        self.assertEqual(user.email, self.user_admin.email)

    def test_get_logout_url(self):
        """Проверка метода get_logout_url."""
        request = self.factory.get('/')
        logout_url = self.backend.get_logout_url(request)
        self.assertEqual(logout_url, self.logout_url)

    def test_process_logout(self):
        """Проверка выхода из системы."""

        request = self.factory.post('/', {
            'username': self.user_admin.username,
            'password': self.admin_password
        })
        self.apply_middleware(request)

        # Входим в систему
        user = self.backend.process_login(request)
        self.assertIsInstance(user, User)
        self.assertTrue(is_authenticated(request))

        # Выходим из системы
        self.backend.process_logout(request)
        self.assertFalse(is_authenticated(request))

    def test_from_config(self):
        """Проверка создания бэкенда из настроек."""
        backend = DjangoBackend.from_config(Config({
            'name': 'django',
            'config': self.config
        }))

        self.assertTrue(backend)
        self.assertIsInstance(backend, DjangoBackend)
        self.assertEqual(backend.name, 'django')
        self.assertIs(backend.checker, DefaultChecker)

        # Проверка создания с указанием другого чекера
        backend = DjangoBackend.from_config(Config({
            'name': 'django',
            'config': self.config,
            'checker_cls': '.'.join((
                CustomChecker.__module__,
                CustomChecker.__name__
            ))
        }))

        self.assertIs(backend.checker, CustomChecker)

    @staticmethod
    def apply_middleware(request):
        SessionMiddleware().process_request(request)
        AuthenticationMiddleware().process_request(request)
