# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from base64 import b64encode
from collections import MutableSequence
from datetime import date
from os import path

import six
from lxml import etree
from django.conf import settings
from django.test.client import RequestFactory
from django.test.testcases import TestCase
from onelogin.saml2.auth import OneLogin_Saml2_Auth as SAMLAuth
from onelogin.saml2.utils import OneLogin_Saml2_Utils as SAMLUtils

from m3_auth.backends.base.remote.user import BaseRemoteUser
from m3_auth.backends.saml2.protocol import SAMLProtocol
from m3_auth.config import Config
from m3_auth.exceptions import AuthError
from m3_auth.exceptions import BadRequest
from m3_auth.fields import DateField
from m3_auth.fields import SimpleField


class TestSAMLUser(BaseRemoteUser):

    """Тестовый класс представляющий информацию о пользователе."""

    surname = SimpleField('surname')
    first_name = SimpleField('first_name')
    patronymic = SimpleField('patronymic')
    date_of_birth = DateField('birth_date', '%d-%m-%Y')

    email = SimpleField('person@mail.ru')

    def __init__(self, data):
        for key, value in data.items():
            if isinstance(value, MutableSequence):
                data[key] = value[0]

        super(TestSAMLUser, self).__init__(data)


class TestSAMLProtocol(TestCase):

    """Проверка класса SAMLProtocol."""

    def setUp(self):
        super(TestSAMLProtocol, self).setUp()
        self.name_id = '492882615acf31c8096b627245d76ae53036c090'
        self.config = Config({
            "custom_base_path": path.join(settings.BASE_DIR, 'conf', 'saml'),
            "logout_redirect_url": "http://localhost:8000/process_logout"
        })
        self.protocol = SAMLProtocol(TestSAMLUser, self.config)
        self.factory = RequestFactory()
        self.http_params = dict(
            secure=True, SERVER_PORT=8000, HTTP_HOST='localhost'
        )

    def create_get(self, url='/', **params):
        post_params = dict(self.http_params)
        post_params.update(params)
        return self.factory.get(url, **post_params)

    def create_post(self, url='/', **params):
        post_params = dict(self.http_params)
        post_params.update(params)
        return self.factory.post(url, **post_params)

    def get_saml_request(self, name, deflate=False):
        """Вощвращает SAMLResponse из файлов."""
        with open(path.join(settings.DATA_PATH, 'saml2', name), 'rb') as fs:
            result = fs.read()

        if deflate:
            return SAMLUtils.deflate_and_base64_encode(result)

        result = b64encode(result)
        return result.encode('utf-8') if six.PY2 else result.decode('utf-8')

    def test_get_login_url(self):
        """Проверка метода get_login_url."""
        request = self.create_post('get_login_url')
        login_url = self.protocol.get_login_url(request)

        self.assertTrue(login_url)

        login_request = self.create_post(login_url)

        self.assertIn('SAMLRequest', login_request.GET)
        self.assertIn('Signature', login_request.GET)
        self.assertIn('RelayState', login_request.GET)
        self.assertIn('SigAlg', login_request.GET)

        self.assertEqual(
            login_request.GET['RelayState'],
            'https://localhost:8000/get_login_url'
        )

        request_xml = SAMLUtils.decode_base64_and_inflate(
            login_request.GET['SAMLRequest'])

        xml = etree.fromstring(request_xml)

        self.assertTrue(request)
        self.assertTrue(len(xml))

    def test_process_login(self):
        """Проверка метода process_login."""

        # Проверка корректного запроса
        response = self.get_saml_request('valid_authn_response.xml')
        request = self.create_post('/', data={'SAMLResponse': response})
        user = self.protocol.process_login(request)

        self.assertIsInstance(user, TestSAMLUser)
        self.assertEqual(user.user_id, self.name_id)
        self.assertEqual(user.surname, 'Иванов')
        self.assertEqual(user.first_name, 'Иван')
        self.assertEqual(user.patronymic, 'Иванович')
        self.assertEqual(user.date_of_birth, date(1990, 1, 1))

        # Проверка некорректного запроса
        response = self.get_saml_request('invalid_authn_response.xml')
        request = self.create_post('/', data={'SAMLResponse': response})
        self.assertRaises(AuthError, self.protocol.process_login, request)

        # Проверка запроса методом GET (который не поддерживается)
        response = self.get_saml_request('invalid_authn_response.xml')
        request = self.factory.get(
            '/', data={'SAMLResponse': response}, HTTP_HOST='localhost')
        self.assertRaises(AuthError, self.protocol.process_login, request)

    def test_get_logout_url(self):
        """Проверка метода get_logout_url."""
        request = self.create_post('localhost:8000/get_logout_url')
        logout_url = self.protocol.get_logout_url(request)

        self.assertTrue(logout_url)

        logout_request = self.create_post(logout_url)

        self.assertIn('SAMLRequest', logout_request.GET)
        self.assertIn('Signature', logout_request.GET)
        self.assertIn('RelayState', logout_request.GET)
        self.assertIn('SigAlg', logout_request.GET)

        self.assertEqual(
            logout_request.GET['RelayState'],
            'http://localhost:8000/process_logout'
        )

        request_xml = SAMLUtils.decode_base64_and_inflate(
            logout_request.GET['SAMLRequest'])

        xml = etree.fromstring(request_xml)

        self.assertTrue(request)
        self.assertTrue(len(xml))

    def test_process_logout(self):
        """Проверка метода process_logout."""
        # Проверка корректного запроса на выход.
        response = self.get_saml_request('valid_logout_response.xml', True)
        request = self.create_get(data={'SAMLResponse': response})
        self.assertTrue(self.protocol.process_logout(request))

        # Проверка некорректного запроса на выход.
        response = self.get_saml_request('invalid_logout_response.xml', True)
        request = self.create_get(data={'SAMLResponse': response})
        self.assertRaises(AuthError, self.protocol.process_logout, request)

    def test_process_single_logout(self):
        """Проверка метода process_single_logout."""

        # Проверка корректного запроса
        saml_request = self.get_saml_request('valid_logout_request.xml', True)
        request = self.create_get(data={'SAMLRequest': saml_request})
        user_id, response_url = self.protocol.process_single_logout(request)

        self.assertEqual(user_id, self.name_id)
        self.assertTrue(response_url)

        response = self.factory.get(response_url)
        self.assertIn('SAMLResponse', response.GET)

        # Проверка некорректного запроса
        saml_request = self.get_saml_request(
            'invalid_logout_request.xml', True)

        request = self.create_get(data={'SAMLRequest': saml_request})

        self.assertRaises(
            AuthError, self.protocol.process_single_logout, request)

        request = self.create_get()
        self.assertRaises(
            BadRequest, self.protocol.process_single_logout, request)

    def test_get_auth(self):
        """Проверка метода _get_auth."""
        request = self.create_post()
        auth = self.protocol._get_auth(request)

        self.assertTrue(auth)
        self.assertIsInstance(auth, SAMLAuth)

    def test_get_request_data(self):
        """Проверка метода _get_request_data."""
        request = self.create_post()
        request_params = self.protocol._get_request_data(request)
        self.assertTrue(request_params)

    def test_from_config(self):
        """Проверка метода from_config."""
        config = Config({
            'user_cls': '.'.join((__name__, TestSAMLUser.__name__)),
            'config': self.config
        })

        protocol = SAMLProtocol.from_config(config)
        self.assertTrue(protocol)
        self.assertIsInstance(protocol, SAMLProtocol)
        self.assertIs(protocol.user_cls, TestSAMLUser)
