# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

import types

from django.contrib.auth.models import User
from django.contrib.auth.middleware import AuthenticationMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.test.client import RequestFactory
from django.test.testcases import TestCase

from m3_auth.backends.django import DjangoBackend
from m3_auth.exceptions import BadRequest
from m3_auth.utils import is_authenticated
from m3_auth.manager import AuthManager
from m3_auth.proxy import SessionProxy


class ManagerTestCase(TestCase):

    """Проверка чекера по умолчанию - DefaultChecker"""

    def setUp(self):
        super(ManagerTestCase, self).setUp()
        self.factory = RequestFactory()
        self.login_url = 'http://localhost:8000/auth/process_login'
        self.logout_url = 'http://localhost:8000/auth/process_logout'

        self.manager = AuthManager()
        self.admin_password = 'admin'
        self.user_admin = User.objects.get(username='admin')

    @staticmethod
    def apply_middleware(request):
        SessionMiddleware().process_request(request)
        AuthenticationMiddleware().process_request(request)

    def test_manager_single_tone(self):
        """Проверка менеджера на то что экземпляр только 1."""
        self.assertEqual(id(self.manager), id(AuthManager()))

    def test_get_backend(self):
        """Проверка метода get_backend."""

        request = self.factory.post('/')
        self.apply_middleware(request)
        # Должно быть исключение, если наименование бэкенда не передано.
        self.assertRaises(BadRequest, self.manager.get_backend, request)

        # Проверка получения бэкенда django
        request = self.factory.post('/', {'backend': 'django'})
        self.apply_middleware(request)

        backend = self.manager.get_backend(request)
        self.assertTrue(backend)
        self.assertIsInstance(backend, DjangoBackend)

        # Проверка приоритета выбора бэкенда, когда пользователь не
        # аутентифицирован

        session_proxy = SessionProxy(request)
        session_proxy.backend = 'unknown_backend'

        backend = self.manager.get_backend(request)
        self.assertTrue(backend)
        self.assertIsInstance(backend, DjangoBackend)

        # Проверка приоритета выбора бэкенда, когда пользователь
        # аутентифицирован

        request = self.factory.post('/', {
            'backend': 'django',
            'username': self.user_admin.username,
            'password': self.admin_password
        })
        self.apply_middleware(request)
        user = self.manager.process_login(request)

        self.assertTrue(user)
        self.assertTrue(is_authenticated(request))

        new_request = self.factory.post('/', {'backend': 'unknown_backend'})
        new_request.session = request.session
        new_request.user = request.user

        backend = self.manager.get_backend(request)
        self.assertEqual(backend.name, 'django')

    def test_get_login_url(self):
        """Проверка получения URL адреса на вход."""
        request = self.factory.post('/', {'backend': 'django'})
        self.apply_middleware(request)
        self.assertEqual(self.manager.get_login_url(request), self.login_url)

    def test_get_logout_url(self):
        """Проверка получения URL адреса на выход."""
        request = self.factory.post('/', {'backend': 'django'})
        self.apply_middleware(request)
        self.assertEqual(self.manager.get_logout_url(request), self.logout_url)

    def test_process_login_and_logout(self):
        """Проверка входа."""
        request = self.factory.post('/', {
            'backend': 'django',
            'username': self.user_admin.username,
            'password': self.admin_password
        })
        self.apply_middleware(request)
        user = self.manager.process_login(request)

        self.assertTrue(user)
        self.assertTrue(is_authenticated(request))

        self.manager.process_logout(request)
        self.assertFalse(is_authenticated(request))

    def test_call_method_of_backend(self):
        """Проверка делегирования вызова неопределенного метода в менеджере 
        - бэкенду"""

        request = self.factory.post('/', {'backend': 'django'})
        self.apply_middleware(request)

        backend = self.manager.get_backend(request)
        backend.test_method = types.MethodType(lambda s, r: True, backend)

        self.assertTrue(self.manager.test_method(request))

    def test_call_undefined_method(self):
        """Проверка вызова не определенного метода."""
        request = self.factory.post('/', {'backend': 'django'})
        self.apply_middleware(request)

        self.assertRaises(BadRequest, self.manager.name, request)
        self.assertRaises(AttributeError, self.manager.undefined, request)
