# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

import six
from django.test.client import RequestFactory
from django.test.testcases import TestCase

from m3_auth.utils import SingletonMeta
from m3_auth.utils import add_query_params
from m3_auth.utils import filter_params
from m3_auth.utils import import_object
from m3_auth.utils import get_request_params
from m3_auth.utils import check_dependency


class SingletonClassTest(six.with_metaclass(SingletonMeta)):

    """Класс для проверки корректной работы метакласса SingletonMeta."""


class TestClassToImport(object):

    """Класс для проверки импорта."""


class UtilsTestCase(TestCase):

    """Проверка модуля utils."""

    def setUp(self):
        super(UtilsTestCase, self).setUp()
        self.factory = RequestFactory()

    def test_singleton_meta(self):
        """Проверка метакласса SingletonMeta."""
        first_instance = SingletonClassTest()
        self.assertEqual(id(first_instance), id(SingletonClassTest()))
        self.assertEqual(id(first_instance), id(SingletonClassTest()))

    def test_import_object(self):
        """Проверка функции import_object."""
        test_class_to_import = import_object('.'.join((
            __name__, TestClassToImport.__name__
        )))
        self.assertIs(test_class_to_import, TestClassToImport)

        # Проверка на импорт не существующего объекта
        self.assertRaises(AttributeError, import_object, '.'.join((
            __name__, 'undefined_object'
        )))

    def test_add_query_params(self):
        """Проверка функции add_query_params."""
        base_url = 'http://localhost:8000/'

        # Без параметров
        new_url = add_query_params(base_url, {})
        self.assertEqual(new_url, base_url)

        # Добавление одного параметра
        new_url = add_query_params(base_url, {'param': 'value'})
        self.assertEqual(new_url, '{}?param=value'.format(base_url))

        # Добавление нескольких параметром
        new_url = add_query_params(base_url, {'attr': 'str', 'param': 'value'})
        self.assertEqual(new_url, '{}?attr=str&param=value'.format(base_url))

        url_with_params = 'http://localhost.ru?some_param=1'
        new_url = add_query_params(url_with_params, {'attr': 'str'})
        self.assertEqual(new_url, '{}&attr=str'.format(url_with_params))

    def test_filter_params(self):
        """Проверка функции filter_params."""

        params = {'a': 1, 'b': 2, 'c': 3}
        self.assertEqual(filter_params(params, []), {})
        self.assertEqual(filter_params(params, ['a']), {'a': 1})
        self.assertEqual(filter_params(params, ['b', 'c']), {'b': 2, 'c': 3})

    def test_get_request_params(self):
        """Проверка функции get_request_params."""
        get_request = self.factory.get('/?param=value')
        post_request = self.factory.post('/', {'post_param': 'post_value'})

        get_params = get_request_params(get_request)
        post_params = get_request_params(post_request)

        self.assertEqual(get_params['param'], 'value')
        self.assertEqual(post_params['post_param'], 'post_value')

    def test_check_dependency(self):
        """Проверка функции check_dependency."""

        self.assertTrue(check_dependency('python3-saml', '1.10.1.1'))
        self.assertFalse(check_dependency('python3-saml', '1.10.1.2'))
        self.assertFalse(check_dependency('undefined-packages', '1.6'))
